/*
 * spindown.c
 *
 * Version: 1.0
 * Date:    25-Feb-2007
 *
 * (c) 2007 Vladimir Grouzdev (vladimir.grouzdev@free.fr)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/* The spindown program stops spinning a SCSI disk if it's idle for a while.
 * The spindown program has been written for Linux 2.4.x.
 *
 * Usage:
 *
 *    spindown [-d] [<time>] [{+|-}<dev>] [{+|-}<dev>] ...
 *
 * Where:
 *
 *    -d specifies that the spindown program will be running as daemon
 *
 *    <time> specifies the disk idle time out in seconds (ex: 1200)
 *
 *    <dev> specifies a SCSI disk device (ex: /dev/sda)
 *
 *	  - The "+" prefix (or no prefix) means that the disk is managed
 *          by the spindown program (positive device).
 *	  - The "-" prefix means that the disk isn't managed by the spindown
 *          program (negative device).
 *	  - If a positive device is specified, the spindown program manages
 *	    only SCSI disks explicitly given in the command line, otherwise
 *	    all detected SCSI disks are managed except negative devices
 *	    explicitly given in the command line.
 *
 * Examples:
 *
 * # spindown -d
 *	- spin down all SCSI disks
 *      - the default idle time out is 10 minutes
 *	- run spindown as daemon
 *
 * # spindown 1200 /dev/sdb
 *	- spin down the /dev/sdb SCSI disk if it's idle for 20 minutes
 *
 * # spindown -d +/dev/sdb
 * 	- spin down the /dev/sdb SCSI disk if it's idle for 10 minutes
 *	- run spindown as daemon
 *
 * # spindown -d 1800 /dev/sda /dev/sdb
 * 	- spin down the /dev/sda and /dev/sdb SCSI disks
 *	- the idle timeout is set to 30 minutes
 *	- run spindown as daemon
 *
 * # spindown 300 -/dev/sda
 *	- spin down all SCSI disks except /dev/sda
 *      - the idle time out is set to 5 minutes
 * 
 * The spindown program periodically looks at I/O disk activity statistics
 * available in /proc/stat. A disk idle condition is detected
 * if the disk statistics aren't changed for a while. Once a disk
 * idle time out is reached, the program sends a SCSI motor control command
 * in order to spin down the disk motor.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>

#define	MIN_IDLE_TIME	10 	// minimal idle time is 10 seconds
#define	DEF_IDLE_TIME	600 	// default idle time is 10 minutes

#define	IDLE_TICK	MIN_IDLE_TIME

#define	SCSI_DEV_SIZE	9
#define	STAT_BUF_SIZE	512

static char _scsi_disk_name[SCSI_DEV_SIZE] = "/dev/sda";

typedef struct scsi_disk_t {			// disk descriptor
    struct scsi_disk_t* next;			// next disk in the list
    int                 id;			// disk ID
    int		        managed;		// disk is managed
    unsigned int	idle;			// disk idle time (in ticks)
    int			stopped;		// disk is stopped
    char	        dev [SCSI_DEV_SIZE];	// disk device name
    char                stat[STAT_BUF_SIZE];	// last disk I/O statistics
} scsi_disk_t;

static scsi_disk_t* _scsi_disks;
static int          _idle;	// idle time (in ticks)
static int          _argc;	// number of <dev> arguments
static char**       _argv;	// table of <dev> arguments

static struct {
    int  inlen;
    int  outlen;
    char cmd[256];
} _scsi_spindown_cmd = {0, 0, {0x1b, 0, 0, 0, 0, 0, }};

    static int
_spindown (char* dev)	// spindown a SCSI disk
{
    int fd, res;
    if ((fd = open(dev, 0)) < 0) {
	fprintf(stderr, "spindown: unable to open %s\n", dev);
	return 0;
    }
    res = ioctl(fd, 1, (void*)&_scsi_spindown_cmd);
    close(fd);
    if (res) {
	fprintf(stderr, "spindown: motor control fails %s\n", dev);
	return 0;		
    }
    return 1;
}

    static int
_is_disk_stat (char* stat)		// is it disk statistics ?
{
    return !strncmp(stat, "disk_io: ", 9);
}

    static char*
_disk_stat_get (char* stat, int* size, int* major, int* id)
{
    char* limit;

    while (*stat && (*stat != '(')) stat++;

    if (!*stat) {
	return 0;
    }

    limit = stat + 1;

    *major = strtol(limit, &limit, 0);

    if (*limit++ != ',') {
	return 0;
    }

    *id = strtol(limit, &limit, 0);

    if ((limit[0] != ')') || (limit[1] != ':') || (limit[2] != '(')) {
	return 0;
    }
    limit += 3;    

    while (*limit && (*limit != ')')) limit++;

    if (!*limit) {
	return 0;        
    }

    limit[1] = 0;

    *size = limit - stat + 2;

    return stat;
}

    static char*
_scsi_stat_get (char* stat, int* size, int* id)
{
    int major;
    while ((stat = _disk_stat_get(stat, size, &major, id))) {
	if (major == 8) {
	    return stat;
	}
	stat += *size;
    }
    return 0;
}

    static int
_scsi_disk_managed (scsi_disk_t* disk)
{
    int i;
    int def = 1;
    if (!_argc) {
	return def;
    }
    for (i = 0; i < _argc; i++) {
	char* dev = _argv[i];
	if (dev[0] == '-') {
	    if (!strcmp(disk->dev, dev+1)) {
	        return 0;
	    }
	} else {
	    def= 0;
	    if (dev[0] == '+') {
		dev++;
	    }
	    if (!strcmp(disk->dev, dev)) {
	        return 1;
	    }
	}
    }
    return def;
}

    static void
_scsi_disk_register (int id, char* stat)	// register a SCSI disk
{
    scsi_disk_t* disk = malloc(sizeof(scsi_disk_t));
    if (!disk) {
	fprintf(stderr, "spindown: malloc() fails\n");
	return;
    }

    disk->id = id;				// disk ID

    strcpy(disk->dev, _scsi_disk_name);
    disk->dev[7] += id;				// disk name

    strcpy(disk->stat, stat);			// current stats

    disk->managed = _scsi_disk_managed(disk);	// enabled/disabled

    disk->stopped = 0;				// disk is spinning
    disk->idle    = 0;				// disk isn't idle

    disk->next  = _scsi_disks;			// add to the list
    _scsi_disks = disk;

    fprintf(stderr, "spindown: %s SCSI disk detected%s\n",
	    disk->dev, (disk->managed ? "" : " (skipped)"));
}

    static scsi_disk_t*
_scsi_disk_find (int id)	// find disk descriptor for a given stat
{
    scsi_disk_t* disk = _scsi_disks;
    while (disk && (disk->id != id)) {
	disk = disk->next;
    }
    return disk;
}

    static void
_scsi_disk_spindown (scsi_disk_t* disk, char* stat) // spindown disk if idle
{
    if (!strcmp(disk->stat, stat)) {
	disk->idle++;
    } else {
	strcpy(disk->stat, stat);
	disk->stopped = 0;
	disk->idle    = 0;
    }
    if ((disk->idle >= _idle) && !disk->stopped) {
	if (_spindown(disk->dev)) {
	    disk->stopped = 1;
	} else {
	    fprintf(stderr, "spindown: unable to spindown %s\n", disk->dev);
	}
    }
}

    static void
_scsi_disks_spindown ()
{
    FILE* file;
    char  buff[STAT_BUF_SIZE];

    if (!(file = fopen("/proc/stat", "r"))) {
	fprintf(stderr, "spindown: unable to open /proc/stat\n");
	return;
    }

    while (fgets(buff, sizeof(buff), file)) {
	if (_is_disk_stat(buff)) {
	    int   size;
	    int   id;
	    char* stat = buff;
	    while ((stat = _scsi_stat_get(stat, &size, &id))) {
	        scsi_disk_t* disk = _scsi_disk_find(id);
	        if (!disk) {
		    _scsi_disk_register(id, stat);
	        } else {
		    if (disk->managed) {
		        _scsi_disk_spindown(disk, stat);
		    }
	        }
		stat += size;
	    }
	}
    }

    fclose(file);
}

    int
main (int argc, char **argv)
{
    unsigned int idle;
    int          daemonize = 0;

    argc--;
    argv++;

    if (argc && !strcmp(argv[0], "-d")) {	// run as daemon
        argc--;
        argv++;
	daemonize = 1;
    }

    if (!argc) {
	idle = DEF_IDLE_TIME;
    } else {
	char* opt = argv[0];
	if (('0' <= opt[0]) && (opt[0] <= '9')) {	// idle time
	    idle = atoi(opt);
	    if (idle < MIN_IDLE_TIME) {
	        idle = MIN_IDLE_TIME;
	    }
	    argc--;
	    argv++;
	} else {
	    idle = DEF_IDLE_TIME;
	}
    }

    _idle = idle / IDLE_TICK;
    _argc = argc;
    _argv = argv;

    if (daemonize && !daemon(0,0)) {
	fprintf(stderr, "spindown: unable to daemonize\n");
    }

    for (;;) {
	_scsi_disks_spindown();
	sleep(IDLE_TICK);
    }

    return 0;
}
